from unittest import TestCase

import pytest

from ......did.did_key import DIDKey
from ......messaging.models.base import BaseModelError
from ......wallet.key_type import ED25519
from .....didcomm_prefix import DIDCommPrefix
from .....didexchange.v1_0.message_types import ARIES_PROTOCOL as DIDEX_1_1
from .....didexchange.v1_0.message_types import DIDEX_1_0
from .....didexchange.v1_0.messages.request import DIDXRequest
from .. import invitation as test_module
from ..invitation import HSProto, InvitationMessage, InvitationMessageSchema
from ..service import Service

TEST_DID = "55GkHamhTU1ZbTbV2ab9DE"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
DID_COMM = "did-communication"


class TestHSProto(TestCase):
    """Test handshake protocol enum."""

    def test_get(self):
        assert HSProto.get(HSProto.RFC160) is HSProto.RFC160
        assert HSProto.get("Old") is HSProto.RFC160
        assert (
            HSProto.get(DIDCommPrefix.qualify_current("connections/1.0"))
            is HSProto.RFC160
        )
        assert HSProto.get(DIDEX_1_0) is HSProto.RFC23
        assert HSProto.get("did-exchange") is HSProto.RFC23
        assert HSProto.get("RFC-23") is HSProto.RFC23
        assert HSProto.get(DIDEX_1_1) is HSProto.DIDEX_1_1
        assert HSProto.get("didexchange/1.1") is HSProto.DIDEX_1_1
        assert HSProto.get("no such protocol") is None
        assert HSProto.get(None) is None


class TestInvitationMessage(TestCase):
    def test_init(self):
        """Test initialization message."""
        invi_msg = InvitationMessage(
            label="A label",
            handshake_protocols=[DIDCommPrefix.qualify_current(DIDEX_1_1)],
            services=[TEST_DID],
        )
        assert invi_msg.services == [TEST_DID]
        assert "out-of-band/1.1/invitation" in invi_msg._type

        service = Service(_id="#inline", _type=DID_COMM, did=TEST_DID)
        invi_msg = InvitationMessage(
            label="A label",
            handshake_protocols=[DIDCommPrefix.qualify_current(DIDEX_1_1)],
            services=[service],
            version="1.0",
        )
        assert invi_msg.services == [service]
        assert "out-of-band/1.0/invitation" in invi_msg._type

    def test_wrap_serde(self):
        """Test conversion of aries message to attachment decorator."""
        msg = {"aries": "message"}
        deco = InvitationMessage.wrap_message(msg)
        assert deco.ident == "request-0"

        obj_x = {"label": "label", "services": ["sample-did"]}
        with pytest.raises(BaseModelError):
            InvitationMessage.deserialize(obj_x)

        invi_schema = InvitationMessageSchema()
        with pytest.raises(test_module.ValidationError):
            invi_schema.validate_fields(obj_x)

        service = Service(
            _id="#inline",
            _type=DID_COMM,
            recipient_keys=[DIDKey.from_public_key_b58(TEST_VERKEY, ED25519).did],
            service_endpoint="http://1.2.3.4:8080/service",
        )
        data_deser = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "services": [{"a": service.serialize()}],
        }
        assert "services" in data_deser

        data_ser = invi_schema.post_dump(data_deser)
        assert "services" in data_ser

        service = Service(_id="#inline", _type=DID_COMM, did=TEST_DID)
        data_deser = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "services": [TEST_DID],
        }
        assert "services" in data_deser

        data_ser = invi_schema.post_dump(data_deser)
        assert "services" in data_ser

    def test_url_round_trip(self):
        service = Service(
            _id="#inline",
            _type=DID_COMM,
            recipient_keys=[
                DIDKey.from_public_key_b58(TEST_VERKEY, ED25519).did,
                DIDKey.from_public_key_b58(TEST_VERKEY, ED25519).key_id,
            ],
            service_endpoint="http://1.2.3.4:8080/service",
        )
        invi_msg = InvitationMessage(
            label="A label",
            handshake_protocols=[DIDCommPrefix.qualify_current(DIDEX_1_1)],
            services=[service],
        )

        url = invi_msg.to_url()
        assert isinstance(url, str)
        invi_msg_rebuilt = InvitationMessage.from_url(url)
        assert isinstance(invi_msg_rebuilt, InvitationMessage)
        assert invi_msg_rebuilt.serialize() == invi_msg.serialize()

    def test_from_no_url(self):
        url = "http://aries.ca/no_ci"
        assert InvitationMessage.from_url(url) is None

    def test_invalid_invi_wrong_type_services(self):
        msg = {"aries": "message"}
        deco = InvitationMessage.wrap_message(msg)
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "services": [123],
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs and "services" in errs

    def test_assign_msg_type_version_to_model_inst(self):
        test_msg = InvitationMessage()
        assert "1.1" in test_msg._type
        assert "1.1" in InvitationMessage.Meta.message_type
        test_msg = InvitationMessage(version="1.2")
        assert "1.2" in test_msg._type
        assert "1.1" in InvitationMessage.Meta.message_type
        test_req = DIDXRequest()
        assert "1.1" in test_req._type
        assert "1.2" in test_msg._type
        assert "1.1" in InvitationMessage.Meta.message_type

    def test_goal_and_code_model(self):
        test_msg = InvitationMessage()
        assert test_msg.goal is None
        assert test_msg.goal_code is None
        test_msg = InvitationMessage(goal="pass test", goal_code="pytest.pass")
        assert test_msg.goal == "pass test"
        assert test_msg.goal_code == "pytest.pass"

    def test_mutual_inclusive_goal_and_goal_code(self):
        msg = {"aries": "message"}
        deco = InvitationMessage.wrap_message(msg)

        # populate both goal and goal_code - good
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "goal_code": "pytest.pass",
            "goal": "pass test",
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs == {}

        # do not include goal and goal_code - good
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs == {}

        # no value for both goal and goal_code - good
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "goal_code": "",
            "goal": "",
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs == {}

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs == {}

        # populate only goal - bad
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "goal": "pass test",
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs and len(errs)
        assert errs and errs["_schema"]
        assert len(errs["_schema"]) == 1
        assert "goal" in errs["_schema"][0]

        # populate only goal_code - bad
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "goal_code": "pytest.pass",
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert errs and len(errs)
        assert errs and errs["_schema"]
        assert len(errs["_schema"]) == 1
        assert "goal_code" in errs["_schema"][0]

        # goal and goal code cannot be set to null/none
        obj_x = {
            "label": "label",
            "requests~attach": [deco.serialize()],
            "goal_code": None,
            "goal": None,
        }

        errs = InvitationMessageSchema().validate(obj_x)
        assert len(errs) == 2
        assert errs["goal"]
        assert errs["goal_code"]



from unittest import IsolatedAsyncioTestCase

from .....admin.request_context import AdminRequestContext
from .....connections.models.conn_record import ConnRecord
from .....protocols.issue_credential.v2_0.models.cred_ex_record import V20CredExRecord
from .....tests import mock
from .....utils.testing import create_test_profile
from .....vc.ld_proofs.error import LinkedDataProofException
from .. import routes as test_module
from ..formats.indy.handler import IndyCredFormatHandler
from ..formats.ld_proof.handler import LDProofCredFormatHandler
from ..messages.cred_format import V20CredFormat
from . import LD_PROOF_VC_DETAIL, TEST_DID


class TestV20CredRoutes(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.session_inject = {}
        self.profile = await create_test_profile(
            settings={
                "admin.admin_api_key": "secret-key",
            }
        )
        self.context = AdminRequestContext.test_context(self.session_inject, self.profile)
        self.request_dict = {
            "context": self.context,
            "outbound_message_router": mock.CoroutineMock(),
        }
        self.request = mock.MagicMock(
            app={},
            match_info={},
            query={},
            __getitem__=lambda _, k: self.request_dict[k],
            headers={"x-api-key": "secret-key"},
        )

    async def test_validate_cred_filter_schema(self):
        schema = test_module.V20CredFilterSchema()
        schema.validate_fields({"indy": {"issuer_did": TEST_DID}})
        schema.validate_fields(
            {"indy": {"issuer_did": TEST_DID, "schema_version": "1.0"}}
        )
        schema.validate_fields(
            {
                "indy": {"issuer_did": TEST_DID},
                "ld_proof": {"credential": {}, "options": {}},
            }
        )
        schema.validate_fields(
            {
                "indy": {},
                "ld_proof": {"credential": {}, "options": {}},
            }
        )
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields(["hopeless", "stop"])
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"veres-one": {"no": "support"}})

    async def test_validate_cred_filter_anoncreds_schema(self):
        schema = test_module.V20CredFilterSchema()
        schema.validate_fields({"anoncreds": {"issuer_id": TEST_DID}})
        schema.validate_fields(
            {"anoncreds": {"issuer_id": TEST_DID, "schema_version": "1.0"}}
        )
        schema.validate_fields(
            {
                "anoncreds": {"issuer_id": TEST_DID},
            }
        )
        schema.validate_fields(
            {
                "anoncreds": {},
            }
        )
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({})
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields(["hopeless", "stop"])
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"veres-one": {"no": "support"}})

    async def test_validate_create_schema(self):
        schema = test_module.V20IssueCredSchemaCore()
        schema.validate(
            {
                "filter": {"indy": {"issuer_did": TEST_DID}},
                "credential_preview": {"..": ".."},
            }
        )
        schema.validate({"filter": {"ld_proof": {"..": ".."}}})

        with self.assertRaises(test_module.ValidationError):
            schema.validate({"filter": {"indy": {"..": ".."}}})

    async def test_validate_bound_offer_request_schema(self):
        schema = test_module.V20CredBoundOfferRequestSchema()
        schema.validate_fields({})
        schema.validate_fields(
            {"filter_": {"indy": {"issuer_did": TEST_DID}}, "counter_preview": {}}
        )
        schema.validate_fields(
            {"filter_": {"ld_proof": {"issuer_did": TEST_DID}}, "counter_preview": {}}
        )
        with self.assertRaises(test_module.ValidationError):
            schema.validate_fields({"filter_": {"indy": {"issuer_did": TEST_DID}}})
            schema.validate_fields({"filter_": {"ld_proof": {"issuer_did": TEST_DID}}})
            schema.validate_fields({"counter_preview": {}})

    async def test_credential_exchange_list(self):
        self.request.query = {
            "thread_id": "dummy",
            "connection_id": "dummy",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.query = mock.CoroutineMock(
                return_value=[
                    V20CredExRecord(
                        connection_id="conn-123",
                        by_format=V20CredFormat.Format.INDY,
                        thread_id="conn-123",
                        cred_ex_id="dummy",
                    )
                ]
            )
            mock_cx_rec.serialize = mock.MagicMock(return_value={"hello": "world"})

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.credential_exchange_list(self.request)
                mock_response.assert_called()

    async def test_credential_exchange_list_x(self):
        self.request.query = {
            "thread_id": "dummy",
            "connection_id": "dummy",
            "role": "dummy",
            "state": "dummy",
        }

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.query = mock.CoroutineMock(side_effect=test_module.StorageError())
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_list(self.request)

    async def test_credential_exchange_retrieve(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value = mock_cx_rec
            mock_cx_rec.serialize = mock.MagicMock()
            mock_cx_rec.serialize.return_value = {"hello": "world"}

            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    mock.MagicMock(  # anoncreds
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),
                    None,  # indy
                    None,  # ld_proof
                    None,  # vc_di
                ]
            )

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.credential_exchange_retrieve(self.request)
                mock_response.assert_called_once_with(
                    {
                        "cred_ex_record": mock_cx_rec.serialize.return_value,
                        "anoncreds": {"...": "..."},
                        "indy": None,
                        "ld_proof": None,
                        "vc_di": None,
                    }
                )

    async def test_credential_exchange_retrieve_indy_ld_proof(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value = mock_cx_rec
            mock_cx_rec.serialize = mock.MagicMock()
            mock_cx_rec.serialize.return_value = {"hello": "world"}

            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    mock.MagicMock(  # anoncreds
                        serialize=mock.MagicMock(return_value={"anon": "creds"})
                    ),
                    mock.MagicMock(  # indy
                        serialize=mock.MagicMock(return_value={"in": "dy"})
                    ),
                    mock.MagicMock(  # ld_proof
                        serialize=mock.MagicMock(return_value={"ld": "proof"})
                    ),
                    mock.MagicMock(  # vc_di
                        serialize=mock.MagicMock(return_value={"vc": "di"})
                    ),
                ]
            )

            with mock.patch.object(test_module.web, "json_response") as mock_response:
                await test_module.credential_exchange_retrieve(self.request)
                mock_response.assert_called_once_with(
                    {
                        "cred_ex_record": mock_cx_rec.serialize.return_value,
                        "anoncreds": {"anon": "creds"},
                        "indy": {"in": "dy"},
                        "ld_proof": {"ld": "proof"},
                        "vc_di": {"vc": "di"},
                    }
                )

    async def test_credential_exchange_retrieve_not_found(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )
            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_retrieve(self.request)

    async def test_credential_exchange_retrieve_x(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value = mock_cx_rec
            mock_cx_rec.serialize = mock.MagicMock(
                side_effect=test_module.BaseModelError()
            )

            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                return_value=None
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_retrieve(self.request)

    async def test_credential_exchange_create(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_cx_rec = mock.MagicMock()
            mock_cred_offer = mock.MagicMock()

            mock_cred_mgr.return_value.prepare_send.return_value = (
                mock_cx_rec,
                mock_cred_offer,
            )

            await test_module.credential_exchange_create(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_create_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_cred_mgr.return_value.prepare_send.side_effect = (
                test_module.StorageError()
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_create(self.request)

    async def test_credential_exchange_create_no_filter(self):
        connection_id = "connection-id"

        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": connection_id,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_create(self.request)
        assert "Missing filter" in str(context.exception)

    async def test_credential_exchange_send(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.CoroutineMock(),
                mock.CoroutineMock(),
            )

            mock_cx_rec = mock.MagicMock()
            mock_cred_offer = mock.MagicMock()

            mock_cred_mgr.return_value.prepare_send.return_value = (
                mock_cx_rec,
                mock_cred_offer,
            )

            await test_module.credential_exchange_send(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_request_no_conn_no_holder_did(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "OobRecord", autospec=True) as mock_oob_rec,
            mock.patch.object(
                test_module, "default_did_from_verkey", autospec=True
            ) as mock_default_did_from_verkey,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_oob_rec.retrieve_by_tag_filter = mock.CoroutineMock(
                return_value=mock.MagicMock(our_recipient_key="our-recipient_key")
            )
            mock_default_did_from_verkey.return_value = "holder-did"

            mock_cred_ex.retrieve_by_id = mock.CoroutineMock()
            mock_cred_ex.retrieve_by_id.return_value.state = (
                mock_cred_ex.STATE_OFFER_RECEIVED
            )
            mock_cred_ex.retrieve_by_id.return_value.connection_id = None

            mock_cred_ex_record = mock.MagicMock()

            mock_cred_mgr.return_value.create_request.return_value = (
                mock_cred_ex_record,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_bound_request(self.request)

            mock_cred_mgr.return_value.create_request.assert_called_once_with(
                mock_cred_ex.retrieve_by_id.return_value, "holder-did"
            )
            mock_response.assert_called_once_with(
                mock_cred_ex_record.serialize.return_value
            )
            mock_default_did_from_verkey.assert_called_once_with("our-recipient_key")

    async def test_credential_exchange_send_no_conn_record(self):
        connection_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": connection_id,
                "credential_preview": preview_spec,
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_not_ready(self):
        connection_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": connection_id,
                "credential_preview": preview_spec,
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_x(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
        ):
            mock_cred_ex_record = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_offer = mock.MagicMock()
            mock_cred_mgr.return_value = mock.CoroutineMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock.CoroutineMock(),
                        mock.CoroutineMock(),
                    )
                ),
                prepare_send=mock.CoroutineMock(
                    return_value=(
                        mock_cred_ex_record,
                        mock_cred_offer,
                    )
                ),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send(self.request)

    async def test_credential_exchange_send_proposal(self):
        connection_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": connection_id,
                "credential_preview": preview_spec,
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cx_rec = mock.MagicMock()
            mock_cred_mgr.return_value.create_proposal.return_value = mock_cx_rec

            await test_module.credential_exchange_send_proposal(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

            self.request["outbound_message_router"].assert_awaited_once()

    async def test_credential_exchange_send_proposal_no_filter(self):
        connection_id = "connection-id"
        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "comment",
                "connection_id": connection_id,
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_send_proposal(self.request)
        assert "Missing filter" in str(context.exception)

    async def test_credential_exchange_send_proposal_no_conn_record(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_proposal.return_value = mock.MagicMock()

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_send_proposal_x(self):
        connection_id = "connection-id"
        preview_spec = {"attributes": [{"name": "attr", "value": "value"}]}

        self.request.json = mock.CoroutineMock(
            return_value={
                "connection_id": connection_id,
                "credential_preview": preview_spec,
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            mock_cx_rec = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_mgr.return_value.create_proposal.return_value = mock_cx_rec

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_send_proposal_not_ready(self):
        self.request.json = mock.CoroutineMock()

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.V20CredPreview, "deserialize", autospec=True),
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_proposal.return_value = mock.MagicMock()

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_proposal(self.request)

    async def test_credential_exchange_create_free_offer(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        self.context.update_settings({"debug.auto_respond_credential_offer": True})
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cx_rec = mock.MagicMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_create_free_offer(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_create_free_offer_no_filter(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_create_free_offer(self.request)
        assert "Missing filter" in str(context.exception)

    async def test_credential_exchange_create_free_offer_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            mock_cx_rec = mock.MagicMock(
                serialize=mock.MagicMock(
                    side_effect=test_module.BaseModelError(),
                ),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_mgr.return_value = mock.MagicMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock_cx_rec,
                        mock.MagicMock(),
                    )
                )
            )
            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_create_free_offer(self.request)

    async def test_credential_exchange_send_free_offer(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cx_rec = mock.MagicMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_free_offer(self.request)
            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_free_offer_vcdi(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "filter": {"vc_di": {"schema_version": "1.0"}},
                "comment": "This is a test comment.",
                "auto_issue": True,
                "auto-remove": True,
                "replacement_id": "test_replacement_id",
                "credential_preview": {
                    "@type": "https://didcomm.org/issue-credential/2.0/credential-preview",
                    "attributes": [
                        {"name": "name", "value": "Alice Smith"},
                        {"name": "date", "value": "2018-05-28"},
                        {"name": "degree", "value": "Maths"},
                        {"name": "birthdate_dateint", "value": "20000330"},
                        {"name": "timestamp", "value": "1711836271"},
                    ],
                },
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            # Mock the creation of a credential offer, especially for handling VC-DI
            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cx_rec = mock.MagicMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            # Call the function you are testing
            await test_module.credential_exchange_send_free_offer(self.request)

            # Validate that the response is correctly structured and called once
            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_free_offer_no_filter(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "comment": "comment",
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
            }
        )

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_send_free_offer(self.request)
        assert "Missing filter" in str(context.exception)

    async def test_credential_exchange_send_free_offer_no_conn_record(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_not_ready(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": True,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_free_offer_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "auto_issue": False,
                "credential_preview": {
                    "attributes": [{"name": "hello", "value": "world"}]
                },
                "filter": {"indy": {"schema_version": "1.0"}},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cx_rec = mock.MagicMock(
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cred_mgr.return_value = mock.CoroutineMock(
                create_offer=mock.CoroutineMock(
                    return_value=(
                        mock_cx_rec,
                        mock.MagicMock(),
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_free_offer(self.request)

    async def test_credential_exchange_send_bound_offer(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_PROPOSAL_RECEIVED
            )

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()

            mock_cx_rec = mock.MagicMock()

            mock_cred_mgr.return_value.create_offer.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_bound_offer(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_bound_offer_linked_data_error(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_PROPOSAL_RECEIVED
            )

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()

            exception_message = "ex"
            mock_cred_mgr.return_value.create_offer.side_effect = (
                LinkedDataProofException(exception_message)
            )
            with self.assertRaises(test_module.web.HTTPBadRequest) as error:
                await test_module.credential_exchange_send_bound_offer(self.request)

            assert exception_message in str(error.exception)

    async def test_credential_exchange_send_bound_offer_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_no_conn_record(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cx_rec.STATE_PROPOSAL_RECEIVED,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_bad_state(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cx_rec.STATE_DONE,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_bound_offer_not_ready(self):
        self.request.json = mock.CoroutineMock(return_value={})
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_PROPOSAL_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_bound_offer(self.request)

    async def test_credential_exchange_send_request(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_OFFER_RECEIVED
            )

            mock_cx_rec = mock.MagicMock()

            mock_cred_mgr.return_value.create_request.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_bound_request(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_request_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_send_bound_request(self.request)

    async def test_credential_exchange_send_request_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value = mock.MagicMock(
                state=mock_cx_rec.STATE_OFFER_RECEIVED,
                save_error_state=mock.CoroutineMock(),
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_bound_request(self.request)

    async def test_credential_exchange_send_request_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_OFFER_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_offer = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_offer.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_bound_request(self.request)

    async def test_credential_exchange_send_free_request(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "filter": {"ld_proof": LD_PROOF_VC_DETAIL},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value.create_request = mock.CoroutineMock()

            mock_cx_rec = mock.MagicMock()

            mock_cred_mgr.return_value.create_request.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_send_free_request(self.request)

            mock_response.assert_called_once_with(mock_cx_rec.serialize.return_value)

    async def test_credential_exchange_send_free_request_no_filter(self):
        self.request.json = mock.CoroutineMock(return_value={"comment": "comment"})

        with self.assertRaises(test_module.web.HTTPBadRequest) as context:
            await test_module.credential_exchange_send_free_request(self.request)
        assert "Missing filter" in str(context.exception)

    async def test_credential_exchange_send_free_request_no_conn_record(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "filter": {"ld_proof": LD_PROOF_VC_DETAIL},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.create_request = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_request.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_send_free_request(self.request)

    async def test_credential_exchange_send_free_request_not_ready(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "filter": {"ld_proof": LD_PROOF_VC_DETAIL},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
        ):
            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.create_request = mock.CoroutineMock()
            mock_cred_mgr.return_value.create_request.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_send_free_request(self.request)

    async def test_credential_exchange_send_free_request_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={
                "filter": {"ld_proof": LD_PROOF_VC_DETAIL},
            }
        )

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response"),
        ):
            mock_cred_mgr.return_value.create_request = mock.CoroutineMock(
                side_effect=[
                    test_module.LedgerError(),
                    test_module.StorageError(),
                ]
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):  # ledger error
                await test_module.credential_exchange_send_free_request(self.request)
            with self.assertRaises(test_module.web.HTTPBadRequest):  # storage error
                await test_module.credential_exchange_send_free_request(self.request)

    async def test_credential_exchange_issue(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_REQUEST_RECEIVED
            )
            mock_cx_rec = mock.MagicMock()

            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    mock.MagicMock(  # anoncreds
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),
                    None,
                    None,  # ld_proof
                    None,  # vc_di
                ]
            )

            mock_cred_mgr.return_value.issue_credential.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_issue(self.request)

            mock_response.assert_called_once_with(
                {
                    "cred_ex_record": mock_cx_rec.serialize.return_value,
                    "anoncreds": {"...": "..."},
                    "indy": None,
                    "ld_proof": None,
                    "vc_di": None,
                }
            )

    async def test_credential_exchange_issue_vcdi(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_REQUEST_RECEIVED
            )
            mock_cx_rec = mock.MagicMock()

            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    None,  # anoncreds
                    None,  # indy
                    None,  # ld_proof
                    mock.MagicMock(  # indy
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),  # vc_di
                ]
            )

            mock_cred_mgr.return_value.issue_credential.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_issue(self.request)

            mock_response.assert_called_once_with(
                {
                    "cred_ex_record": mock_cx_rec.serialize.return_value,
                    "anoncreds": None,
                    "indy": None,
                    "ld_proof": None,
                    "vc_di": {"...": "..."},
                }
            )

    async def test_credential_exchange_issue_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cx_rec = mock.MagicMock(
            conn_id="dummy",
            serialize=mock.MagicMock(),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
        ):
            mock_cx_rec.state = mock_cx_rec_cls.STATE_REQUEST_RECEIVED
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock(return_value=mock_cx_rec)

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.issue_credential = mock.CoroutineMock()
            mock_cred_mgr.return_value.issue_credential.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_REQUEST_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            mock_cred_mgr.return_value.issue_credential = mock.CoroutineMock()
            mock_cred_mgr.return_value.issue_credential.return_value = (
                mock.MagicMock(),
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_rev_reg_full_indy(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cx_rec = mock.MagicMock(
            conn_id="dummy",
            serialize=mock.MagicMock(),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
        ):
            mock_cx_rec.state = mock_cx_rec_cls.STATE_REQUEST_RECEIVED
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock(return_value=mock_cx_rec)

            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = True

            mock_issue_cred = mock.CoroutineMock(
                side_effect=test_module.IndyIssuerError()
            )
            mock_cred_mgr.return_value.issue_credential = mock_issue_cred

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_rev_reg_full_anoncreds(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cx_rec = mock.MagicMock(
            conn_id="dummy",
            serialize=mock.MagicMock(),
            save_error_state=mock.CoroutineMock(),
        )
        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
        ):
            mock_cx_rec.state = mock_cx_rec_cls.STATE_REQUEST_RECEIVED
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock(return_value=mock_cx_rec)

            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = True

            mock_issue_cred = mock.AsyncMock(
                side_effect=test_module.AnonCredsIssuerError()
            )
            mock_cred_mgr.return_value.issue_credential = mock_issue_cred

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_issue_deser_x(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        mock_cx_rec = V20CredExRecord(connection_id="dummy", cred_ex_id="dummy")
        mock_cx_rec.serialize = mock.MagicMock(side_effect=test_module.BaseModelError())
        mock_cx_rec.save_error_state = mock.CoroutineMock()

        mock_conn_rec = mock.MagicMock(ConnRecord, autospec=True)
        mock_conn_rec.retrieve_by_id = mock.CoroutineMock(return_value=ConnRecord())

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock(return_value=mock_cx_rec)
            mock_cred_mgr.return_value = mock.MagicMock(
                issue_credential=mock.CoroutineMock(
                    return_value=(
                        mock_cx_rec,
                        mock.MagicMock(),
                    )
                )
            )
            mock_cred_mgr.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    mock.MagicMock(  # indy
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),
                    None,  # ld_proof
                ]
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_issue(self.request)

    async def test_credential_exchange_store(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_CREDENTIAL_RECEIVED
            )
            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    None,  # anoncreds
                    mock.MagicMock(  # indy
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),
                    None,  # ld_proof
                    None,  # vc_di
                ]
            )

            mock_cx_rec = mock.MagicMock()

            mock_cred_mgr.return_value.store_credential.return_value = mock_cx_rec
            mock_cred_mgr.return_value.send_cred_ack.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_store(self.request)

            mock_response.assert_called_once_with(
                {
                    "cred_ex_record": mock_cx_rec.serialize.return_value,
                    "anoncreds": None,
                    "indy": {"...": "..."},
                    "ld_proof": None,
                    "vc_di": None,
                }
            )

    async def test_credential_exchange_store_bad_cred_id_json(self):
        self.request.json = mock.CoroutineMock(
            side_effect=test_module.JSONDecodeError("Nope", "Nope", 0)
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response") as mock_response,
            mock.patch.object(
                LDProofCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_ld_proof_get_detail_record,
            mock.patch.object(
                IndyCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_indy_get_detail_record,
        ):
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec_cls.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_CREDENTIAL_RECEIVED
            )

            mock_cx_rec = V20CredExRecord(connection_id="dummy", cred_ex_id="dummy")
            mock_cx_rec.serialize = mock.MagicMock(
                return_value={"cred_ex_id": "dummy", "state": "credential_received"}
            )
            mock_cx_rec.save_error_state = mock.CoroutineMock()
            mock_cx_rec.connection_id = "dummy"

            mock_conn_rec = mock.MagicMock(ConnRecord, autospec=True)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(return_value=ConnRecord())

            mock_indy_get_detail_record.return_value = mock.MagicMock(  # indy
                serialize=mock.MagicMock(return_value={"...": "..."})
            )
            mock_ld_proof_get_detail_record.return_value = None  # ld_proof

            mock_cred_mgr.return_value.store_credential.return_value = mock_cx_rec
            mock_cred_mgr.return_value.send_cred_ack.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            await test_module.credential_exchange_store(self.request)

            mock_response.assert_called_once_with(
                {
                    "cred_ex_record": mock_cx_rec.serialize.return_value,
                    "anoncreds": None,
                    "indy": {"...": "..."},
                    "ld_proof": None,
                    "vc_di": None,
                }
            )

    async def test_credential_exchange_store_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredExRecord", autospec=True
        ) as mock_cx_rec:
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.side_effect = test_module.StorageNotFoundError()

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_no_conn_record(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    state=mock_cx_rec.STATE_CREDENTIAL_RECEIVED,
                    save_error_state=mock.CoroutineMock(),
                )
            )

            # Emulate storage not found (bad connection id)
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            mock_cred_mgr.return_value.store_credential.return_value = mock_cx_rec
            mock_cred_mgr.return_value.send_cred_ack.return_value = (
                mock_cx_rec,
                mock.MagicMock(),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_not_ready(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True) as mock_conn_rec,
            mock.patch.object(test_module, "V20CredManager", autospec=True),
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec,
        ):
            mock_cx_rec.connection_id = "conn-123"
            mock_cx_rec.thread_id = "conn-123"
            mock_cx_rec.retrieve_by_id = mock.CoroutineMock()
            mock_cx_rec.retrieve_by_id.return_value.state = (
                test_module.V20CredExRecord.STATE_CREDENTIAL_RECEIVED
            )

            # Emulate connection not ready
            mock_conn_rec.retrieve_by_id = mock.CoroutineMock()
            mock_conn_rec.retrieve_by_id.return_value.is_ready = False

            with self.assertRaises(test_module.web.HTTPForbidden):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_store_x(self):
        self.request.json = mock.CoroutineMock()
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "ConnRecord", autospec=True),
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cx_rec_cls,
            mock.patch.object(test_module.web, "json_response"),
            mock.patch.object(V20CredFormat.Format, "handler") as mock_handler,
        ):
            mock_cx_rec = mock.MagicMock(
                state=mock_cx_rec_cls.STATE_CREDENTIAL_RECEIVED,
                serialize=mock.MagicMock(side_effect=test_module.BaseModelError()),
                save_error_state=mock.CoroutineMock(),
            )
            mock_cx_rec_cls.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock()
            )
            mock_handler.return_value.get_detail_record = mock.CoroutineMock(
                side_effect=[
                    None,  # anoncreds
                    mock.MagicMock(  # indy
                        serialize=mock.MagicMock(return_value={"...": "..."})
                    ),
                    None,  # ld_proof
                    None,  # vc_di
                ]
            )

            mock_cred_mgr.return_value = mock.MagicMock(
                store_credential=mock.CoroutineMock(return_value=mock_cx_rec),
                send_cred_ack=mock.CoroutineMock(
                    return_value=(
                        mock_cx_rec,
                        mock.MagicMock(),
                    )
                ),
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_store(self.request)

    async def test_credential_exchange_remove(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(
                test_module, "V20CredManager", autospec=True
            ) as mock_cred_mgr,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_mgr.return_value = mock.MagicMock(
                delete_cred_ex_record=mock.CoroutineMock()
            )
            await test_module.credential_exchange_remove(self.request)

            mock_response.assert_called_once_with({})

    async def test_credential_exchange_remove_bad_cred_ex_id(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value = mock.MagicMock(
                delete_cred_ex_record=mock.CoroutineMock(
                    side_effect=test_module.StorageNotFoundError()
                )
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_remove(self.request)

    async def test_credential_exchange_remove_x(self):
        self.request.match_info = {"cred_ex_id": "dummy"}

        with mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value = mock.MagicMock(
                delete_cred_ex_record=mock.CoroutineMock(
                    side_effect=test_module.StorageError()
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_remove(self.request)

    async def test_credential_exchange_problem_report(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}
        magic_report = mock.MagicMock()

        with (
            mock.patch.object(test_module, "V20CredManager", autospec=True),
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cred_ex,
            mock.patch.object(
                test_module, "problem_report_for_record", mock.MagicMock()
            ) as mock_problem_report,
            mock.patch.object(test_module.web, "json_response") as mock_response,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(save_error_state=mock.CoroutineMock())
            )
            mock_problem_report.return_value = magic_report

            await test_module.credential_exchange_problem_report(self.request)

            self.request["outbound_message_router"].assert_awaited_once_with(
                magic_report,
                connection_id=mock_cred_ex.retrieve_by_id.return_value.connection_id,
            )
            mock_response.assert_called_once_with({})

    async def test_credential_exchange_problem_report_bad_cred_ex_id(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "V20CredManager", autospec=True),
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                side_effect=test_module.StorageNotFoundError()
            )

            with self.assertRaises(test_module.web.HTTPNotFound):
                await test_module.credential_exchange_problem_report(self.request)

    async def test_credential_exchange_problem_report_x(self):
        self.request.json = mock.CoroutineMock(
            return_value={"description": "Did I say no problem? I meant 'no: problem.'"}
        )
        self.request.match_info = {"cred_ex_id": "dummy"}

        with (
            mock.patch.object(test_module, "V20CredManager", autospec=True),
            mock.patch.object(test_module, "problem_report_for_record", mock.MagicMock()),
            mock.patch.object(
                test_module, "V20CredExRecord", autospec=True
            ) as mock_cred_ex,
        ):
            mock_cred_ex.retrieve_by_id = mock.CoroutineMock(
                return_value=mock.MagicMock(
                    save_error_state=mock.CoroutineMock(
                        side_effect=test_module.StorageError()
                    )
                )
            )

            with self.assertRaises(test_module.web.HTTPBadRequest):
                await test_module.credential_exchange_problem_report(self.request)

    async def test_register(self):
        mock_app = mock.MagicMock()
        mock_app.add_routes = mock.MagicMock()

        await test_module.register(mock_app)
        mock_app.add_routes.assert_called_once()

    async def test_post_process_routes(self):
        mock_app = mock.MagicMock(_state={"swagger_dict": {}})
        test_module.post_process_routes(mock_app)
        assert "tags" in mock_app._state["swagger_dict"]



"""DID Key class and resolver methods."""

from typing import List, Optional

from ..vc.ld_proofs.constants import DID_V1_CONTEXT_URL, SECURITY_CONTEXT_MULTIKEY_URL
from ..wallet.crypto import ed25519_pk_to_curve25519
from ..wallet.key_type import (
    BLS12381G1,
    BLS12381G1G2,
    BLS12381G2,
    ED25519,
    P256,
    X25519,
    KeyType,
    KeyTypes,
)
from ..wallet.util import b58_to_bytes, bytes_to_b58


class DIDKey:
    """DID Key parser and resolver."""

    _key_type: KeyType
    _public_key: bytes

    def __init__(self, public_key: bytes, key_type: KeyType) -> None:
        """Initialize new DIDKey instance."""
        self._public_key = public_key
        self._key_type = key_type

    @classmethod
    def from_public_key(cls, public_key: bytes, key_type: KeyType) -> "DIDKey":
        """Initialize new DIDKey instance from public key and key type."""

        return cls(public_key, key_type)

    @classmethod
    def from_public_key_b58(cls, public_key: str, key_type: KeyType) -> "DIDKey":
        """Initialize new DIDKey instance from base58 encoded public key and key type."""
        public_key_bytes = b58_to_bytes(public_key)
        return cls.from_public_key(public_key_bytes, key_type)

    @classmethod
    def from_fingerprint(cls, fingerprint: str, key_types=None) -> "DIDKey":
        """Initialize new DIDKey instance from multibase encoded fingerprint.

        The fingerprint contains both the public key and key type.
        """
        # Assert fingerprint is in multibase format
        assert fingerprint[0] == "z"

        # Get key bytes, remove multicodec prefix
        key_bytes_with_prefix = b58_to_bytes(fingerprint[1:])

        # Get associated key type with prefixed bytes
        if not key_types:
            key_types = KeyTypes()
        key_type = key_types.from_prefixed_bytes(key_bytes_with_prefix)

        if not key_type:
            raise Exception(
                f"No key type for prefixed public key '{key_bytes_with_prefix}' found."
            )

        # Remove the prefix bytes to get the public key
        prefix_len = len(key_type.multicodec_prefix)
        public_key_bytes = key_bytes_with_prefix[prefix_len:]

        return cls(public_key_bytes, key_type)

    @classmethod
    def from_did(cls, did: str) -> "DIDKey":
        """Initialize a new DIDKey instance from a fully qualified did:key string.

        Extracts the fingerprint from the did:key and uses that to construct the did:key.
        """
        did_parts = did.split("#")
        _, fingerprint = did_parts[0].split("did:key:")

        return cls.from_fingerprint(fingerprint)

    @property
    def prefixed_public_key(self) -> bytes:
        """Getter for multicodec prefixed public key."""
        return b"".join([self.key_type.multicodec_prefix, self.public_key])

    @property
    def fingerprint(self) -> str:
        """Getter for did key fingerprint."""
        return f"z{bytes_to_b58(self.prefixed_public_key)}"

    @property
    def did(self) -> str:
        """Getter for full did:key string."""
        return f"did:key:{self.fingerprint}"

    @property
    def did_doc(self) -> dict:
        """Getter for did document associated with did:key."""
        resolver = DID_KEY_RESOLVERS[self.key_type]
        return resolver(self)

    @property
    def public_key(self) -> bytes:
        """Getter for public key."""
        return self._public_key

    @property
    def public_key_b58(self) -> str:
        """Getter for base58 encoded public key."""
        return bytes_to_b58(self.public_key)

    @property
    def key_type(self) -> KeyType:
        """Getter for key type."""
        return self._key_type

    @property
    def key_id(self) -> str:
        """Getter for key id."""
        return f"{self.did}#{self.fingerprint}"


def construct_did_key_bls12381g2(did_key: "DIDKey") -> dict:
    """Construct BLS12381G2 did:key.

    Args:
        did_key (DIDKey): did key instance to parse bls12381g2 did:key document from

    Returns:
        dict: The bls12381g2 did:key did document

    """

    return construct_did_signature_key_base(
        id=did_key.did,
        key_id=did_key.key_id,
        verification_method={
            "id": did_key.key_id,
            "type": "Bls12381G2Key2020",
            "controller": did_key.did,
            "publicKeyBase58": did_key.public_key_b58,
        },
    )


def construct_did_key_bls12381g1(did_key: "DIDKey") -> dict:
    """Construct BLS12381G1 did:key.

    Args:
        did_key (DIDKey): did key instance to parse bls12381g1 did:key document from

    Returns:
        dict: The bls12381g1 did:key did document

    """

    return construct_did_signature_key_base(
        id=did_key.did,
        key_id=did_key.key_id,
        verification_method={
            "id": did_key.key_id,
            "type": "Bls12381G1Key2020",
            "controller": did_key.did,
            "publicKeyBase58": did_key.public_key_b58,
        },
    )


def construct_did_key_bls12381g1g2(did_key: "DIDKey") -> dict:
    """Construct BLS12381G1G2 did:key.

    Args:
        did_key (DIDKey): did key instance to parse bls12381g1g2 did:key document from

    Returns:
        dict: The bls12381g1g2 did:key did document

    """

    g1_public_key = did_key.public_key[:48]
    g2_public_key = did_key.public_key[48:]

    bls12381g1_key = DIDKey.from_public_key(g1_public_key, BLS12381G1)
    bls12381g2_key = DIDKey.from_public_key(g2_public_key, BLS12381G2)

    bls12381g1_key_id = f"{did_key.did}#{bls12381g1_key.fingerprint}"
    bls12381g2_key_id = f"{did_key.did}#{bls12381g2_key.fingerprint}"

    return {
        "@context": DID_V1_CONTEXT_URL,
        "id": did_key.did,
        "verificationMethod": [
            {
                "id": bls12381g1_key_id,
                "type": "Bls12381G1Key2020",
                "controller": did_key.did,
                "publicKeyBase58": bls12381g1_key.public_key_b58,
            },
            {
                "id": bls12381g2_key_id,
                "type": "Bls12381G2Key2020",
                "controller": did_key.did,
                "publicKeyBase58": bls12381g2_key.public_key_b58,
            },
        ],
        "authentication": [bls12381g1_key_id, bls12381g2_key_id],
        "assertionMethod": [bls12381g1_key_id, bls12381g2_key_id],
        "capabilityDelegation": [bls12381g1_key_id, bls12381g2_key_id],
        "capabilityInvocation": [bls12381g1_key_id, bls12381g2_key_id],
        "keyAgreement": [],
    }


def construct_did_key_x25519(did_key: "DIDKey") -> dict:
    """Construct X25519 did:key.

    Args:
        did_key (DIDKey): did key instance to parse x25519 did:key document from

    Returns:
        dict: The x25519 did:key did document

    """

    return {
        "@context": DID_V1_CONTEXT_URL,
        "id": did_key.did,
        "verificationMethod": [
            {
                "id": did_key.key_id,
                "type": "X25519KeyAgreementKey2019",
                "controller": did_key.did,
                "publicKeyBase58": did_key.public_key_b58,
            },
        ],
        "authentication": [],
        "assertionMethod": [],
        "capabilityDelegation": [],
        "capabilityInvocation": [],
        "keyAgreement": [did_key.key_id],
    }


def construct_did_key_ed25519(did_key: "DIDKey") -> dict:
    """Construct Ed25519 did:key.

    Args:
        did_key (DIDKey): did key instance to parse ed25519 did:key document from

    Returns:
        dict: The ed25519 did:key did document

    """
    curve25519 = ed25519_pk_to_curve25519(did_key.public_key)
    x25519 = DIDKey.from_public_key(curve25519, X25519)

    did_doc = construct_did_signature_key_base(
        id=did_key.did,
        key_id=did_key.key_id,
        verification_method={
            "id": did_key.key_id,
            "type": "Ed25519VerificationKey2018",
            "controller": did_key.did,
            "publicKeyBase58": did_key.public_key_b58,
        },
    )

    # Ed25519 has pair with X25519
    did_doc["keyAgreement"].append(
        {
            "id": f"{did_key.did}#{x25519.fingerprint}",
            "type": "X25519KeyAgreementKey2019",
            "controller": did_key.did,
            "publicKeyBase58": bytes_to_b58(curve25519),
        }
    )

    return did_doc


def construct_did_key_p256(did_key: "DIDKey") -> dict:
    """Construct P256 did:key.

    Args:
        did_key (DIDKey): did key instance to parse p256 did:key document from

    Returns:
        dict: The p256 did:key did document

    """

    did_doc = construct_did_signature_key_base(
        id=did_key.did,
        key_id=did_key.key_id,
        verification_method={
            "id": did_key.key_id,
            "type": "Multikey",
            "controller": did_key.did,
            "publicKeyMultibase": did_key.fingerprint,
        },
        extra_context=[SECURITY_CONTEXT_MULTIKEY_URL],
    )

    return did_doc


def construct_did_signature_key_base(
    *,
    id: str,
    key_id: str,
    verification_method: dict,
    extra_context: Optional[List[str]] = None,
):
    """Create base did key structure to use for most signature keys.

    May not be suitable for all did key types

    """

    return {
        "@context": [DID_V1_CONTEXT_URL] + (extra_context or []),
        "id": id,
        "verificationMethod": [verification_method],
        "authentication": [key_id],
        "assertionMethod": [key_id],
        "capabilityDelegation": [key_id],
        "capabilityInvocation": [key_id],
        "keyAgreement": [],
    }


DID_KEY_RESOLVERS = {
    ED25519: construct_did_key_ed25519,
    X25519: construct_did_key_x25519,
    P256: construct_did_key_p256,
    BLS12381G2: construct_did_key_bls12381g2,
    BLS12381G1: construct_did_key_bls12381g1,
    BLS12381G1G2: construct_did_key_bls12381g1g2,
}
